"""
Module implementing transformer-based defences against poisoning attacks.
"""
from art.defences.transformer.poisoning.neural_cleanse import NeuralCleanse
from art.defences.transformer.poisoning.strip import STRIP
